///////////////////////////////////////////////////////////////////////
// MPEG-2 Plugin for VirtualDub 1.8.1+
// Copyright (C) 2007-2008 fccHandler
// Copyright (C) 1998-2008 Avery Lee
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
///////////////////////////////////////////////////////////////////////

// IVDXVideoDecoder and IVDXVideoDecoderModel interfaces

#include <windows.h>

#include "InputFileMPEG2.h"
#include "MPEG2Decoder.h"

#include <crtdbg.h>

#define MPEG_BUFFER_BIDIRECTIONAL	(0)
#define MPEG_BUFFER_BACKWARD		(1)
#define MPEG_BUFFER_FORWARD			(2)
#define MPEG_BUFFER_AUXILIARY		(3)

// Chroma formats
#define CHROMA420		1
#define CHROMA422		2
#define CHROMA444		3

//////////////////////////////////////////////////////////////////////////
//
//
//						VideoDecoderMPEG2
//
//
//////////////////////////////////////////////////////////////////////////

class VideoDecoderMPEG2 : public vdxunknown<IVDXVideoDecoder> {

private:
	InputFileMPEG2 *const parentPtr;
	IMPEG2Decoder *mpDecoder;

	sint64 mPrevFrame;

	void DecodeFrameBuffer(int buffer);

public:
	VideoDecoderMPEG2(InputFileMPEG2 *pp);
	~VideoDecoderMPEG2();

	const void *VDXAPIENTRY DecodeFrame(const void *inputBuffer, uint32 data_len, bool is_preroll, sint64 streamFrame, sint64 targetFrame);
	uint32		VDXAPIENTRY GetDecodePadding();
	void		VDXAPIENTRY Reset();
	bool		VDXAPIENTRY IsFrameBufferValid();
	const VDXPixmap& VDXAPIENTRY GetFrameBuffer();
	bool		VDXAPIENTRY SetTargetFormat(int format, bool useDIBAlignment);
	bool		VDXAPIENTRY SetDecompressedFormat(const VDXBITMAPINFOHEADER *pbih);

	const void *VDXAPIENTRY GetFrameBufferBase();
	bool		VDXAPIENTRY IsDecodable(sint64 sample_num);

protected:
	VDXPixmap mPixmap;
	unsigned char *mFrameBuffer;
};


VideoDecoderMPEG2::VideoDecoderMPEG2(InputFileMPEG2 *const pp)
	: parentPtr(pp)
{
	parentPtr->AddRef();

_RPT0(_CRT_WARN, "VideoDecoderMPEG2 constructed\n");

	mPrevFrame = -1;

	mFrameBuffer = (unsigned char *)malloc(parentPtr->width * parentPtr->height * 4);
	if (mFrameBuffer != NULL) {
		memset(mFrameBuffer, 0, parentPtr->width * parentPtr->height * 4);
	}

	mpDecoder = CreateMPEG2Decoder();
	if (mpDecoder != NULL) {
		mpDecoder->Init(parentPtr->width, parentPtr->height, parentPtr->seq_ext);
	}
}

VideoDecoderMPEG2::~VideoDecoderMPEG2() {
_RPT0(_CRT_WARN, "VideoDecoderMPEG2 destructed\n");
	if (mpDecoder) mpDecoder->Destruct();
	free(mFrameBuffer);
	parentPtr->Release();
}

void VideoDecoderMPEG2::DecodeFrameBuffer(int buffer) {
	const long w = parentPtr->width;
	const long h = parentPtr->height;

	if (mFrameBuffer == NULL) return;

	using namespace nsVDXPixmap;

	switch (mPixmap.format) {

	case kPixFormat_YUV420_Planar:
		mpDecoder->Decode420Planar(mPixmap.data, mPixmap.pitch,
			mPixmap.data2, mPixmap.pitch2, mPixmap.data3,
			mPixmap.pitch3, buffer);
		break;

	case kPixFormat_Y8:
		{
			const unsigned char *src;
			unsigned char *dst;
			long y, srcpitch;

			src = (const unsigned char *)mpDecoder->GetYBuffer(buffer, srcpitch);
			dst = (unsigned char *)mPixmap.data;
			for (y = h; y > 0; --y) {
				memcpy(dst, src, w);
				src += srcpitch;
				dst += mPixmap.pitch;
			}
		}
		break;

	case kPixFormat_YUV422_Planar:
		mpDecoder->Decode422Planar(mPixmap.data, mPixmap.pitch,
			mPixmap.data2, mPixmap.pitch2, mPixmap.data3,
			mPixmap.pitch3, buffer);
		break;

	case kPixFormat_YUV422_UYVY:
		mpDecoder->DecodeUYVY(mPixmap.data, mPixmap.pitch, buffer);
		break;

	case kPixFormat_YUV422_YUYV:
		mpDecoder->DecodeYUYV(mPixmap.data, mPixmap.pitch, buffer);
		break;

	case kPixFormat_XRGB1555:
		mpDecoder->DecodeRGB15(mPixmap.data, mPixmap.pitch, buffer);
		break;

	case kPixFormat_RGB565:
		mpDecoder->DecodeRGB16(mPixmap.data, mPixmap.pitch, buffer);
		break;

	case kPixFormat_RGB888:
		mpDecoder->DecodeRGB24(mPixmap.data, mPixmap.pitch, buffer);
		break;

	case kPixFormat_XRGB8888:
		mpDecoder->DecodeRGB32(mPixmap.data, mPixmap.pitch, buffer);
		break;
	}
}


const void *VideoDecoderMPEG2::DecodeFrame(
	const void *inputBuffer, uint32 data_len,
	bool is_preroll, sint64 streamFrame, sint64 targetFrame)
{
	const MPEGSampleInfo *m1, *m2;
	long f1, f2, sframe;
	int buffer, b1, b2;

//_RPT3(_CRT_WARN, "DecodeFrame(is_preroll = %i, stream = %i, target = %i)\n",
//	(int)is_preroll, (int)streamFrame, (int)targetFrame);

	if (mpDecoder == NULL) return mFrameBuffer;

	sframe = (long)streamFrame;
	if (sframe < 0) sframe = (long)targetFrame;
	if ((unsigned long)sframe >= (unsigned long)parentPtr->fields) return mFrameBuffer;

	// Update decoder acceleration flags
	f1 = (long)parentPtr->mContext.mpCallbacks->GetCPUFeatureFlags();
	if (f1 & kVDXCPUF_MMX) {
		if (f1 & kVDXCPUF_ISSE) {
			mpDecoder->EnableAcceleration(IMPEG2Decoder::eAccel_ISSE);
		} else {
			mpDecoder->EnableAcceleration(IMPEG2Decoder::eAccel_MMX);
		}
	} else {
		mpDecoder->EnableAcceleration(IMPEG2Decoder::eAccel_None);
	}

	// Do we use "matrix coefficients" for YCbCr to RGB?
	mpDecoder->EnableMatrixCoefficients(parentPtr->fAllowMatrix);

	f1 = parentPtr->video_field_map[sframe];

	if (is_preroll) {
		// Just decode, but skip B-frames

		// If this is a laced frame, we actually want the second!
		if ((f1 & 1) && sframe > 0) {
			f2 = f1 >> 1;
			f1 = parentPtr->video_field_map[sframe - 1] >> 1;

			m1 = &(parentPtr->video_sample_list[f1]);
			m2 = &(parentPtr->video_sample_list[f2]);

			buffer = mpDecoder->GetFrameBuffer(f2);

			if (m2->frame_type != MPEG_FRAME_TYPE_B) {
				if (buffer != MPEG_BUFFER_BACKWARD) {
					mpDecoder->SwapFrameBuffers(MPEG_BUFFER_FORWARD, MPEG_BUFFER_BACKWARD);
					if (buffer != MPEG_BUFFER_FORWARD) {
						mpDecoder->DecodeFrame((char *)inputBuffer + m1->size, m2->size, f2, MPEG_BUFFER_BACKWARD, MPEG_BUFFER_FORWARD, MPEG_BUFFER_BACKWARD);
					}
				}
			}

		} else {
			f1 >>= 1;
			m1 = &(parentPtr->video_sample_list[f1]);
			buffer = mpDecoder->GetFrameBuffer(f1);

			if (m1->frame_type != MPEG_FRAME_TYPE_B) {
				if (buffer != MPEG_BUFFER_BACKWARD) {
					mpDecoder->SwapFrameBuffers(MPEG_BUFFER_FORWARD, MPEG_BUFFER_BACKWARD);
					if (buffer != MPEG_BUFFER_FORWARD) {
						mpDecoder->DecodeFrame(inputBuffer, data_len, f1, MPEG_BUFFER_BACKWARD, MPEG_BUFFER_FORWARD, MPEG_BUFFER_BACKWARD);
					}
				}
			}
		}
	}

	if (!is_preroll) {
		// Decode, assemble, and display

		if ((f1 & 1) && sframe > 0) {

			f2 = f1 >> 1;
			f1 = parentPtr->video_field_map[sframe - 1] >> 1;

			m1 = &(parentPtr->video_sample_list[f1]);
			m2 = &(parentPtr->video_sample_list[f2]);

if (inputBuffer != NULL) {

			b1 = mpDecoder->GetFrameBuffer(f1);
			b2 = mpDecoder->GetFrameBuffer(f2);

			if (m1->frame_type != MPEG_FRAME_TYPE_B) {

				if (m2->frame_type != MPEG_FRAME_TYPE_B) {
					// II, IP, PI, PP

					// Do we have f1?
					if (b1 == MPEG_BUFFER_FORWARD) {
						// Maybe we also have f2?
						if (b2 != MPEG_BUFFER_BACKWARD) {
							mpDecoder->DecodeFrame((char *)inputBuffer + m1->size, m2->size, f2, MPEG_BUFFER_BACKWARD, MPEG_BUFFER_FORWARD, MPEG_BUFFER_BACKWARD);
						}
					} else if (b1 == MPEG_BUFFER_BACKWARD) {
						// Hmmm, got f1 in other buffer
						mpDecoder->SwapFrameBuffers(MPEG_BUFFER_FORWARD, MPEG_BUFFER_BACKWARD);
						if (b2 != MPEG_BUFFER_FORWARD) {
							mpDecoder->DecodeFrame((char *)inputBuffer + m1->size, m2->size, f2, MPEG_BUFFER_BACKWARD, MPEG_BUFFER_FORWARD, MPEG_BUFFER_BACKWARD);
						}
					} else if (b2 == MPEG_BUFFER_FORWARD) {
						// Not got f1, but got f2 in other buffer
						mpDecoder->DecodeFrame(inputBuffer, m1->size, f1, MPEG_BUFFER_BACKWARD, MPEG_BUFFER_FORWARD, MPEG_BUFFER_BACKWARD);
						mpDecoder->SwapFrameBuffers(MPEG_BUFFER_FORWARD, MPEG_BUFFER_BACKWARD);
					} else if (b2 == MPEG_BUFFER_BACKWARD) {
						// Not got f1, but got f2
						mpDecoder->DecodeFrame(inputBuffer, m1->size, f1, MPEG_BUFFER_FORWARD, MPEG_BUFFER_BACKWARD, MPEG_BUFFER_FORWARD);
					} else {
						// Ain't got neither
						mpDecoder->DecodeFrame(inputBuffer, m1->size, f1, MPEG_BUFFER_BACKWARD, MPEG_BUFFER_FORWARD, MPEG_BUFFER_BACKWARD);
						mpDecoder->SwapFrameBuffers(MPEG_BUFFER_FORWARD, MPEG_BUFFER_BACKWARD);
						mpDecoder->DecodeFrame((char *)inputBuffer + m1->size, m2->size, f2, MPEG_BUFFER_BACKWARD, MPEG_BUFFER_FORWARD, MPEG_BUFFER_BACKWARD);
					}

				} else {
					// IB, PB
					if (b1 != MPEG_BUFFER_FORWARD) {
						if (b1 == MPEG_BUFFER_BACKWARD) {
							// Hmm, f1 in wrong buffer
							mpDecoder->SwapFrameBuffers(MPEG_BUFFER_FORWARD, MPEG_BUFFER_BACKWARD);
						} else {
							mpDecoder->DecodeFrame(inputBuffer, m1->size, f1, MPEG_BUFFER_BACKWARD, MPEG_BUFFER_FORWARD, MPEG_BUFFER_BACKWARD);
						}
					}
					if (b2 != MPEG_BUFFER_BIDIRECTIONAL) {
						mpDecoder->DecodeFrame((char *)inputBuffer + m1->size, m2->size, f2, MPEG_BUFFER_BIDIRECTIONAL, MPEG_BUFFER_FORWARD, MPEG_BUFFER_BACKWARD);
					}
				}
			} else {
				if (m2->frame_type != MPEG_FRAME_TYPE_B) {
					// BI, BP (degenerate)
					if (b2 != MPEG_BUFFER_BACKWARD) {
						mpDecoder->SwapFrameBuffers(MPEG_BUFFER_FORWARD, MPEG_BUFFER_BACKWARD);
						if (b2 != MPEG_BUFFER_FORWARD) {
							mpDecoder->DecodeFrame((char *)inputBuffer + m1->size, m2->size, f2, MPEG_BUFFER_BACKWARD, MPEG_BUFFER_FORWARD, MPEG_BUFFER_BACKWARD);
						}
					}
					if (b1 != MPEG_BUFFER_BIDIRECTIONAL) {
						mpDecoder->DecodeFrame(inputBuffer, m1->size, f1, MPEG_BUFFER_BIDIRECTIONAL, MPEG_BUFFER_FORWARD, MPEG_BUFFER_BACKWARD);
					}
				} else {
					// BB
					if (b1 != MPEG_BUFFER_BIDIRECTIONAL) {
						mpDecoder->DecodeFrame(inputBuffer, m1->size, f1, MPEG_BUFFER_BIDIRECTIONAL, MPEG_BUFFER_FORWARD, MPEG_BUFFER_BACKWARD);
					}
					if (b2 != MPEG_BUFFER_AUXILIARY) {
						mpDecoder->DecodeFrame((char *)inputBuffer + m1->size, m2->size, f2, MPEG_BUFFER_AUXILIARY, MPEG_BUFFER_FORWARD, MPEG_BUFFER_BACKWARD);
					}
				}
			}

}	// inputBuffer != NULL

			// Now assemble the fields into auxiliary buffer
			b1 = mpDecoder->GetFrameBuffer(f1);
			b2 = mpDecoder->GetFrameBuffer(f2);

			if (m2->frame_type == MPEG_FRAME_TYPE_B) {
				if (m1->frame_type == MPEG_FRAME_TYPE_B) {
					buffer = MPEG_BUFFER_AUXILIARY;
				} else {
					buffer = MPEG_BUFFER_BIDIRECTIONAL;
				}
			} else {
				buffer = MPEG_BUFFER_BACKWARD;
			}

			if ((m1->bitflags & 3) != FRAME_PICTURE) {
				if ((m1->bitflags & 3) == TOP_FIELD) {
					// m1 top + m2 bottom
					mpDecoder->CopyField(MPEG_BUFFER_AUXILIARY, 1, buffer, 1);
					mpDecoder->CopyField(MPEG_BUFFER_AUXILIARY, 0, b1, 0);
				} else {
					// m1 bottom + m2 top
					mpDecoder->CopyField(MPEG_BUFFER_AUXILIARY, 0, buffer, 0);
					mpDecoder->CopyField(MPEG_BUFFER_AUXILIARY, 1, b1, 1);
				}

			} else if ((m2->bitflags & 3) != FRAME_PICTURE) {
				if ((m2->bitflags & 3) == TOP_FIELD) {
					// m1 bottom + m2 top
					mpDecoder->CopyField(MPEG_BUFFER_AUXILIARY, 0, buffer, 0);
					mpDecoder->CopyField(MPEG_BUFFER_AUXILIARY, 1, b1, 1);
				} else {
					// m1 top + m2 bottom
					mpDecoder->CopyField(MPEG_BUFFER_AUXILIARY, 1, buffer, 1);
					mpDecoder->CopyField(MPEG_BUFFER_AUXILIARY, 0, b1, 0);
				}

			} else {
				if (m1->bitflags & PICTUREFLAG_RFF) {
					if (m1->bitflags & PICTUREFLAG_TFF) {
						// m1 top + m2 bottom
						mpDecoder->CopyField(MPEG_BUFFER_AUXILIARY, 1, buffer, 1);
						mpDecoder->CopyField(MPEG_BUFFER_AUXILIARY, 0, b1, 0);
					} else {
						// m1 bottom + m2 top
						mpDecoder->CopyField(MPEG_BUFFER_AUXILIARY, 0, buffer, 0);
						mpDecoder->CopyField(MPEG_BUFFER_AUXILIARY, 1, b1, 1);
					}
				} else {
					if (m1->bitflags & PICTUREFLAG_TFF) {
						// m1 bottom + m2 top
						mpDecoder->CopyField(MPEG_BUFFER_AUXILIARY, 0, buffer, 0);
						mpDecoder->CopyField(MPEG_BUFFER_AUXILIARY, 1, b1, 1);
					} else {
						// m1 top + m2 bottom
						mpDecoder->CopyField(MPEG_BUFFER_AUXILIARY, 1, buffer, 1);
						mpDecoder->CopyField(MPEG_BUFFER_AUXILIARY, 0, b1, 0);
					}
				}
			}

			buffer = MPEG_BUFFER_AUXILIARY;

		} else {
			// Frame picture
			f1 >>= 1;
			m1 = &(parentPtr->video_sample_list[f1]);
			buffer = mpDecoder->GetFrameBuffer(f1);

if (inputBuffer != NULL) {

			if (m1->frame_type != MPEG_FRAME_TYPE_B) {
				if (buffer != MPEG_BUFFER_FORWARD) {
					// Got f1 in back buffer?
					if (buffer != MPEG_BUFFER_BACKWARD) {
						mpDecoder->SwapFrameBuffers(MPEG_BUFFER_FORWARD, MPEG_BUFFER_BACKWARD);
						if (buffer != MPEG_BUFFER_FORWARD) {
							mpDecoder->DecodeFrame(inputBuffer, m1->size, f1, MPEG_BUFFER_BACKWARD, MPEG_BUFFER_FORWARD, MPEG_BUFFER_BACKWARD);
						}
						buffer = MPEG_BUFFER_BACKWARD;
					}
				}
			} else {
				if (buffer != MPEG_BUFFER_BIDIRECTIONAL) {
					mpDecoder->DecodeFrame(inputBuffer, m1->size, f1, MPEG_BUFFER_BIDIRECTIONAL, MPEG_BUFFER_FORWARD, MPEG_BUFFER_BACKWARD);
					buffer = MPEG_BUFFER_BIDIRECTIONAL;
				}
			}

}	// inputBuffer != NULL

		}

		DecodeFrameBuffer(buffer);
		mPrevFrame = sframe;	//targetFrame;
	}

	return mFrameBuffer;
}

uint32 VideoDecoderMPEG2::GetDecodePadding() {
	return 0;
}

void VideoDecoderMPEG2::Reset() {
	if (mpDecoder != NULL) mpDecoder->ClearFrameBuffers();
	mPrevFrame = -1;
}

bool VideoDecoderMPEG2::IsFrameBufferValid() {
	return (mPrevFrame >= 0);
}

const VDXPixmap& VideoDecoderMPEG2::GetFrameBuffer() {
	return mPixmap;
}

bool VideoDecoderMPEG2::SetTargetFormat(int format, bool useDIBAlignment) {
	int w = parentPtr->width;
	int h = parentPtr->height;

	using namespace nsVDXPixmap;

	// Default format:
	// For CHROMA420 this should be YV12
	// For CHROMA422 this should be YUY2
	// For CHROMA444 this should be RGB24

	if (format == kPixFormat_Null) {
		const int chroma_format = (parentPtr->seq_ext >> 9) & 3;

		if ((w & 1) == 0 && chroma_format != CHROMA444) {
			if ((h & 1) == 0 && chroma_format != CHROMA422) {
				format = kPixFormat_YUV420_Planar;
			} else {
				format = kPixFormat_YUV422_YUYV;
			}
		} else {
			format = kPixFormat_RGB888;
		}
	}

	switch (format) {

	case kPixFormat_YUV420_Planar:
		// Width and height must be even
		if ((w | h) & 1) return false;
		mPixmap.data		= mFrameBuffer;
		mPixmap.palette		= NULL;
		mPixmap.format		= format;
		mPixmap.w			= w;
		mPixmap.h			= h;
		mPixmap.pitch		= w;
		mPixmap.data2		= mFrameBuffer + w * h;
		mPixmap.pitch2		= w >> 1;
		mPixmap.data3		= mFrameBuffer + w * h + (w >> 1) * (h >> 1);
		mPixmap.pitch3		= w >> 1;
		break;

	case kPixFormat_Y8:
// Note: according to the SDK, we do not align the pitch
		mPixmap.data		= mFrameBuffer;
		mPixmap.pitch		= w;
		mPixmap.palette		= NULL;
		mPixmap.format		= format;
		mPixmap.w			= w;
		mPixmap.h			= h;
		mPixmap.data2		= NULL;
		mPixmap.pitch2		= 0;
		mPixmap.data3		= NULL;
		mPixmap.pitch3		= 0;
		break;

	case kPixFormat_YUV422_UYVY:
	case kPixFormat_YUV422_YUYV:
		// Width must be even
		if (w & 1) return false;
		mPixmap.data		= mFrameBuffer;
		mPixmap.pitch		= w * 2;
		mPixmap.palette		= NULL;
		mPixmap.format		= format;
		mPixmap.w			= w;
		mPixmap.h			= h;
		mPixmap.data2		= NULL;
		mPixmap.pitch2		= 0;
		mPixmap.data3		= NULL;
		mPixmap.pitch3		= 0;
		break;

	case kPixFormat_XRGB1555:
	case kPixFormat_RGB565:
		mPixmap.pitch		= (w * 2 + 3) & 0xFFFFFFFC;
		mPixmap.data		= mFrameBuffer + (h - 1) * mPixmap.pitch;
		mPixmap.pitch		= -mPixmap.pitch;
		mPixmap.palette		= NULL;
		mPixmap.format		= format;
		mPixmap.w			= w;
		mPixmap.h			= h;
		mPixmap.data2		= NULL;
		mPixmap.pitch2		= 0;
		mPixmap.data3		= NULL;
		mPixmap.pitch3		= 0;
		break;

	case kPixFormat_RGB888:
		mPixmap.pitch		= (w * 3 + 3) & 0xFFFFFFFC;
		mPixmap.data		= mFrameBuffer + (h - 1) * mPixmap.pitch;
		mPixmap.pitch		= -mPixmap.pitch;
		mPixmap.palette		= NULL;
		mPixmap.format		= format;
		mPixmap.w			= w;
		mPixmap.h			= h;
		mPixmap.data2		= NULL;
		mPixmap.pitch2		= 0;
		mPixmap.data3		= NULL;
		mPixmap.pitch3		= 0;
		break;
	
	case kPixFormat_XRGB8888:
		mPixmap.pitch		= w * 4;
		mPixmap.data		= mFrameBuffer + (h - 1) * mPixmap.pitch;
		mPixmap.pitch		= -mPixmap.pitch;
		mPixmap.palette		= NULL;
		mPixmap.format		= format;
		mPixmap.w			= w;
		mPixmap.h			= h;
		mPixmap.data2		= NULL;
		mPixmap.pitch2		= 0;
		mPixmap.data3		= NULL;
		mPixmap.pitch3		= 0;
		break;

	case kPixFormat_YUV422_Planar:
		// Width must be even
		if (w & 1) return false;
		mPixmap.data		= mFrameBuffer;
		mPixmap.pitch		= w;
		mPixmap.palette		= NULL;
		mPixmap.format		= format;
		mPixmap.w			= w;
		mPixmap.h			= h;
		mPixmap.data2		= mFrameBuffer + w * h;
		mPixmap.pitch2		= w >> 1;
		mPixmap.data3		= mFrameBuffer + w * h + (w >> 1) * h;
		mPixmap.pitch3		= w >> 1;
		break;

	default:
		return false;

	}

	Reset();

	return true;
}

bool VideoDecoderMPEG2::SetDecompressedFormat(const VDXBITMAPINFOHEADER *pbih) {
	return false;
}

const void *VideoDecoderMPEG2::GetFrameBufferBase() {
	return mFrameBuffer;
}

bool VideoDecoderMPEG2::IsDecodable(sint64 sample_num64) {
	if (mpDecoder != NULL) {
		return (mpDecoder->GetFrameBuffer((long)sample_num64) >= 0);
	}
	return false;
}


//////////////////////////////////////////////////////////////////////////
//
//
//						VideoDecoderModelMPEG2
//
//
//////////////////////////////////////////////////////////////////////////

class VideoDecoderModelMPEG2 : public vdxunknown<IVDXVideoDecoderModel> {

private:
	InputFileMPEG2 *const parentPtr;

	long frame_forw, frame_back, frame_bidir, frame_aux;

	sint64 mSampleFirst;
	sint64 mSampleLast;

	sint64 stream_current_frame;
	sint64 stream_desired_frame;

	sint64 prev_field_IP(sint64 f64);
	sint64 next_field_IP(sint64 f64);

public:
	VideoDecoderModelMPEG2(InputFileMPEG2 *pp);
	~VideoDecoderModelMPEG2();

	void	VDXAPIENTRY Reset();
	void	VDXAPIENTRY SetDesiredFrame(sint64 frame_num);
	sint64	VDXAPIENTRY GetNextRequiredSample(bool& is_preroll);
	int		VDXAPIENTRY GetRequiredCount();
};


VideoDecoderModelMPEG2::VideoDecoderModelMPEG2(InputFileMPEG2 *const pp)
	: parentPtr(pp)
{
	parentPtr->AddRef();

_RPT0(_CRT_WARN, "VideoDecoderModelMPEG2 constructed\n");

	mSampleFirst = 0;
	mSampleLast = parentPtr->fields;
}

VideoDecoderModelMPEG2::~VideoDecoderModelMPEG2() {
_RPT0(_CRT_WARN, "VideoDecoderModelMPEG2 destructed\n");
	parentPtr->Release();
}

void VideoDecoderModelMPEG2::Reset() {
	frame_forw = frame_back = frame_bidir = frame_aux = -1;

//_RPT0(_CRT_WARN, "Reset()\n");
}

sint64 VideoDecoderModelMPEG2::prev_field_IP(sint64 f64) {

	if (f64 < mSampleLast) {
		const MPEGSampleInfo *m1;
		long f1;

		while (--f64 >= mSampleFirst) {
			f1 = parentPtr->video_field_map[f64] >> 1;
			m1 = &(parentPtr->video_sample_list[f1]);
			if (m1->frame_type != MPEG_FRAME_TYPE_B) return f64;
		}
	}

    return -1;
}

sint64 VideoDecoderModelMPEG2::next_field_IP(sint64 f64) {

	if (f64 >= mSampleFirst) {
		const MPEGSampleInfo *m1;
		long f1;

		while (++f64 < mSampleLast) {
			f1 = parentPtr->video_field_map[f64] >> 1;
			m1 = &(parentPtr->video_sample_list[f1]);
			if (m1->frame_type != MPEG_FRAME_TYPE_B) return f64;
		}
	}

    return -1;
}

void VideoDecoderModelMPEG2::SetDesiredFrame(sint64 frame_num) {
	const MPEGSampleInfo *m1;
	long f1, f2, prev, next;
	sint64 n;

	stream_desired_frame = stream_current_frame = frame_num;

//_RPT1(_CRT_WARN, "SetDesiredFrame(%i)\n", (int)frame_num);

	f1 = parentPtr->video_field_map[frame_num];

	// If the desired frame is laced... (aargh)

	if ((f1 & 1) && stream_current_frame > mSampleFirst) {
		const MPEGSampleInfo *m2;

		f2 = f1 >> 1;
		f1 = parentPtr->video_field_map[stream_current_frame - 1] >> 1;

		m1 = &(parentPtr->video_sample_list[f1]);
		m2 = &(parentPtr->video_sample_list[f2]);

		// There are a lot of possibilities here!

		// Just remember, the question we are answering
		// is whether we need to move stream_current_frame
		// from its original position at stream_desired_frame

		if (m1->frame_type == MPEG_FRAME_TYPE_I) {
			// II or IP, do nothing

			if (m2->frame_type == MPEG_FRAME_TYPE_B) {
				// IB

				// Decode f1, then future IP, then f2

// Fix for Clip.m2v?  We need not be concerned with I, only B:
//				if (frame_forw != f1 || frame_bidir != f2) {
				if (frame_bidir != f2) {

					// Find the future reference
					n = next_field_IP(stream_current_frame);
					if (n >= 0) {
						next = parentPtr->video_field_map[n] >> 1;

						if (frame_forw != f1 || frame_back != next) {
							if (frame_back == f1) {
								// Start at next reference
								stream_current_frame = n;
							} else {
								// Start at I frame
								--stream_current_frame;
							}
						}
					}
				}
			}

		} else if (m1->frame_type == MPEG_FRAME_TYPE_P) {

			if (m2->frame_type != MPEG_FRAME_TYPE_B) {
				// PI, or PP

				// Decode previous reference, then f1, then f2

				// If we don't have f1, back to prev I/P
				// Otherwise start from f2

				if ((frame_forw != f1 || frame_back != f2)
						&& frame_back != f1) {

					// Start from f1 if possible
					--stream_current_frame;
					while (stream_current_frame > mSampleFirst) {
						n = prev_field_IP(stream_current_frame);
						if (n < 0) break;	// ouch
						prev = parentPtr->video_field_map[n] >> 1;
						if (frame_back == prev) break;
						stream_current_frame = n;
						if (parentPtr->video_sample_list[prev].frame_type == MPEG_FRAME_TYPE_I) break;
					}
				}

			} else {
				// PB

				// Decode previous reference, then f1,
				// then future reference, then f2

				if (frame_forw != f1 && frame_back != f1) {

					// Missing the P, go back to prev I/P
					--stream_current_frame;
					while (stream_current_frame > mSampleFirst) {
						n = prev_field_IP(stream_current_frame);
						prev = parentPtr->video_field_map[n] >> 1;
						if (frame_back == prev) break;
						stream_current_frame = n;
						if (parentPtr->video_sample_list[prev].frame_type == MPEG_FRAME_TYPE_I) break;
					}

				} else if (frame_bidir != f2) {
					// Got P, but missing B

					// We'll need the future reference
					n = next_field_IP(stream_current_frame);
					if (n >= 0) {
						next = parentPtr->video_field_map[n] >> 1;

					// If P is in forward buffer but we don't have
					// next, start at P.  (Not sure that can happen)

					// If P is in back buffer and we don't have next
					// then start at next.
						if (frame_forw == f1 && frame_back != next) {
							// Something got screwed
							--stream_current_frame;
							while (stream_current_frame > mSampleFirst) {
								n = prev_field_IP(stream_current_frame);
								prev = parentPtr->video_field_map[n] >> 1;
								if (frame_back == prev) break;
								stream_current_frame = n;
								if (parentPtr->video_sample_list[prev].frame_type == MPEG_FRAME_TYPE_I) break;
							}
						} else if (frame_back == f1) {
							stream_current_frame = n;
						}
					}
				}
			}

		} else {
			// BI, BP, or BB

			if (m2->frame_type != MPEG_FRAME_TYPE_B) {
				// BI, or BP

				// Fix for VdubCrash2.mpg
				if (frame_bidir != f1 || (frame_forw != f2 && frame_back != f2 && m2->frame_type != MPEG_FRAME_TYPE_I)) {

					// Do we have the previous reference?
					n = prev_field_IP(stream_current_frame);
					if (n >= 0) {
						prev = parentPtr->video_field_map[n] >> 1;

						// If we don't have it, start there
						if (frame_back != prev) {

							stream_current_frame = n;
							while (stream_current_frame > mSampleFirst) {
								n = prev_field_IP(stream_current_frame);
								prev = parentPtr->video_field_map[n] >> 1;
								if (frame_back == prev) break;
								stream_current_frame = n;
								if (parentPtr->video_sample_list[prev].frame_type == MPEG_FRAME_TYPE_I) break;
							}

						}
					} else {
						// Degenerate; no previous reference
						--stream_current_frame;
					}
				}

			} else {
				// BB (special case)

				if (frame_bidir != f1 || frame_aux != f2) {
					// We're going to need to look for 2 refs
					sint64 n2;

					// Find the references
					n = prev_field_IP(stream_current_frame);
					n2 = next_field_IP(stream_current_frame);

					if (n >= 0 && n2 >= 0) {

						prev = parentPtr->video_field_map[n] >> 1;
						next = parentPtr->video_field_map[n2] >> 1;

						if (frame_forw == prev && frame_back == next) {
							// Wow, we have both references!
							// Start from f1 if we don't have it

//							if (frame_bidir != f1) {
//							// (So far, only seen when moving backwards)
//								--stream_current_frame;
//							}

						} else if (frame_back == prev) {
							// We have the prev reference, start at next
							stream_current_frame = n2;

						} else {
							// Drat, go back to prev I/P

							// (This can happen on a random seek)
							stream_current_frame = n;
							while (stream_current_frame > mSampleFirst) {
								n = prev_field_IP(stream_current_frame);
								prev = parentPtr->video_field_map[n] >> 1;
								if (frame_back == prev) break;
								stream_current_frame = n;
								if (parentPtr->video_sample_list[prev].frame_type == MPEG_FRAME_TYPE_I) break;
							}
						}
					}
				}
			}
		}
		return;
	}


	// No lacing involved, decode as usual

	f1 >>= 1;
	m1 = &(parentPtr->video_sample_list[f1]);

	switch(m1->frame_type) {

	case MPEG_FRAME_TYPE_P:
		if (frame_back != f1 && frame_forw != f1) {
			// Back to previous I/P
			while (stream_current_frame > mSampleFirst) {
				n = prev_field_IP(stream_current_frame);
				if (n < 0) break;	// ouch
				f1 = parentPtr->video_field_map[n] >> 1;
				if (frame_back == f1) break;
				stream_current_frame = n;
				if (parentPtr->video_sample_list[f1].frame_type == MPEG_FRAME_TYPE_I) break;
			}
		}
		break;

	case MPEG_FRAME_TYPE_B:

		if (frame_bidir != f1) {
			sint64 np, nn;

			// Back to previous I/P

			// Find the past reference
			np = prev_field_IP(stream_current_frame);
			if (np >= 0) {
				prev = parentPtr->video_field_map[np] >> 1;
			} else {
				prev = 0;
			}

			// Find the future reference
			nn = next_field_IP(stream_current_frame);
			if (nn >= 0) {
				next = parentPtr->video_field_map[nn] >> 1;
			} else {
				next = parentPtr->vframes - 1;
			}

			if (np >= 0 && nn >= 0) {

				// We only need to move if we're missing one of these
				if (frame_forw != prev || frame_back != next) {
					if (frame_back != prev) {
						// Back to prev I/P
						stream_current_frame = np;

						while (stream_current_frame > mSampleFirst) {
							n = prev_field_IP(stream_current_frame);
							if (n < 0) break;	// ouch
							f1 = parentPtr->video_field_map[n] >> 1;
							if (frame_back == f1) break;
							stream_current_frame = n;
							if (parentPtr->video_sample_list[f1].frame_type == MPEG_FRAME_TYPE_I) break;
						}
					} else {
						// Already got prev, start at next
						stream_current_frame = nn;
					}
				}

			}
		}
	}
}

sint64 VideoDecoderModelMPEG2::GetNextRequiredSample(bool& is_preroll) {
	const MPEGSampleInfo *m1, *m2;
	long f1, f2;
	sint64 current;

	current = stream_current_frame;

//	_RPT2(_CRT_WARN, "GetNextRequiredSample(current = %i, desired = %i)\n",
//		(int)current, (int)stream_desired_frame);

	f1 = parentPtr->video_field_map[current];

	// Only decode I/P frames, until we reach the desired frame

	if (current < stream_desired_frame) {
		f1 >>= 1;
		if (parentPtr->video_sample_list[f1].frame_type != MPEG_FRAME_TYPE_B) {
			if (frame_back != f1) {
				frame_forw = frame_back;
				frame_back = f1;
			}
		}
		stream_current_frame = next_field_IP(current);

		// Watch out for this weird situation:
		if (stream_current_frame == (stream_desired_frame - 1)) {
			f1 = parentPtr->video_field_map[stream_current_frame];
			f2 = parentPtr->video_field_map[stream_desired_frame];
			if ((f1 & 1) == 1 && f1 == (f2 + 1)) {
				// We don't need f1, so skip to desired
				stream_current_frame = stream_desired_frame;
			}
		}

//_RPT1(_CRT_WARN, "GetNextRequiredSample returning %i, True\n", (int)current);

		is_preroll = true;
		return current;
	}

	if (current > stream_desired_frame) {
		// We're one frame away from being finished
		f1 >>= 1;
		if (parentPtr->video_sample_list[f1].frame_type != MPEG_FRAME_TYPE_B) {
			if (frame_back != f1) {
				frame_forw = frame_back;
				frame_back = f1;
			}
		}
		stream_current_frame = stream_desired_frame;

//_RPT1(_CRT_WARN, "GetNextRequiredSample returning %i, True\n", (int)current);

		is_preroll = true;
		return current;
	}

	// We've reached the desired frame.  But in the
	// case of a laced frame, we must emulate everything
	// streamGetFrame does to ensure it can be assembled.

	if ((f1 & 1) && current > mSampleFirst) {
		f2 = f1 >> 1;
		f1 = parentPtr->video_field_map[current - 1] >> 1;

		m1 = &(parentPtr->video_sample_list[f1]);
		m2 = &(parentPtr->video_sample_list[f2]);

		if (m1->frame_type != MPEG_FRAME_TYPE_B) {

			if (m2->frame_type != MPEG_FRAME_TYPE_B) {
				// II, IP, PI, PP

				// Do we have f1?
				if (frame_forw == f1) {
					if (frame_back == f2) goto Neg1;
					frame_back = f2;
				} else if (frame_back == f1) {
					// Hmmm, got f1 in wrong buffer
					frame_back = frame_forw;
					frame_forw = f1;
					if (frame_back == f2) goto Neg1;
					frame_back = f2;
				} else if (frame_forw == f2) {
					// Not got f1, but got f2 in wrong buffer
					frame_forw = f1;
					frame_back = f2;
				} else if (frame_back == f2) {
					// Huh? Not got f1, but f2 in correct buffer!
					// (streamGetFrame will take care of this)
					frame_forw = f1;
				} else {
					// Ain't got neither f1 nor f2 (triple negative)
					frame_forw = f1;
					frame_back = f2;
				}

			} else {
				// IB or PB
				if (frame_forw != f1) {
					if (frame_back == f1) {
						// Hmm, f1 in wrong buffer
						frame_back = frame_forw;
						frame_forw = f1;
					} else {
						frame_back = f1;
					}
				}
				if (frame_bidir == f2) goto Neg1;
				frame_bidir = f2;
			}
		} else {
			if (m2->frame_type != MPEG_FRAME_TYPE_B) {
				// BI or BP
				if (frame_back != f2) {
					frame_forw = frame_back;
					frame_back = f2;
				}
				if (frame_bidir == f1) goto Neg1;
				frame_bidir = f1;
			} else {
				// BB (special case)
				if (frame_bidir == f1 && frame_aux == f2) goto Neg1;
				frame_bidir = f1;
				frame_aux = f2;
			}
		}


	} else {
		// No lacing, thank goodness
		f1 >>= 1;
		m1 = &(parentPtr->video_sample_list[f1]);

		if (m1->frame_type != MPEG_FRAME_TYPE_B) {
			if (frame_back == f1) goto Neg1;
//			if (frame_forw == f1) goto Neg1;
			frame_forw = frame_back;
			frame_back = f1;
		} else {
			if (frame_bidir == f1) goto Neg1;
			frame_bidir = f1;
		}
	}

//_RPT1(_CRT_WARN, "GetNextRequiredSample returning %i, False\n", (int)current);

	is_preroll = false;
	return current;

Neg1:
	// Finished!

//_RPT0(_CRT_WARN, "GetNextRequiredSample returning -1, False\n");

	frame_aux = -1;
	is_preroll = false;
	return -1;
}

int VideoDecoderModelMPEG2::GetRequiredCount() {

// Return int = the number of frames which need to be decoded

// Perhaps the easiest way to do this is to emulate
// VirtualDub calls to GetNextRequiredSample.

// First, save ALL tracking variables...

	sint64 saved_stream_current_frame = stream_current_frame;
	sint64 saved_stream_desired_frame = stream_desired_frame;
	long saved_frame_forw = frame_forw;
	long saved_frame_back = frame_back;
	long saved_frame_bidir = frame_bidir;
	long saved_frame_aux = frame_aux;

	int needed = 1;

	for (;;) {
		bool p;
		sint64 next = GetNextRequiredSample(p);
		if (next == -1) break;

		long f1 = parentPtr->video_field_map[next];

		if ((f1 & 1) && next > mSampleFirst) {
			f1 = parentPtr->video_field_map[next - 1] >> 1;
		}
		++needed;
	}

	stream_current_frame = saved_stream_current_frame;
	stream_desired_frame = saved_stream_desired_frame;
	frame_forw = saved_frame_forw;
	frame_back = saved_frame_back;
	frame_bidir = saved_frame_bidir;
	frame_aux = saved_frame_aux;

//_RPT1(_CRT_WARN, "GetRequiredCount returning %i\n", needed);

	return needed;
}

IVDXVideoDecoder *CreateVideoDecoderMPEG2(InputFileMPEG2 *const pp) {
	return new VideoDecoderMPEG2(pp);
}

IVDXVideoDecoderModel *CreateVideoDecoderModelMPEG2(InputFileMPEG2 *const pp) {
	return new VideoDecoderModelMPEG2(pp);
}
