///////////////////////////////////////////////////////////////////////
// MPEG-2 Plugin for VirtualDub 1.8.1+
// Copyright (C) 2007-2008 fccHandler
// Copyright (C) 1998-2008 Avery Lee
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
///////////////////////////////////////////////////////////////////////

// MPEG System, Video, and Audio Parsers

#include <windows.h>
#include <commctrl.h>
#include <stdio.h>

#include "parser.h"
#include "audio.h"

#include "resource.h"

#include <crtdbg.h>

extern HMODULE g_hModule;

#define VIDPKT_TYPE_SEQUENCE_START		0x1B3
#define VIDPKT_TYPE_SEQUENCE_END		0x1B7
#define VIDPKT_TYPE_GROUP_START			0x1B8
#define VIDPKT_TYPE_PICTURE_START		0x100
#define VIDPKT_TYPE_SLICE_START_MIN		0x101
#define VIDPKT_TYPE_SLICE_START_MAX		0x1AF
#define VIDPKT_TYPE_EXT_START			0x1B5
#define VIDPKT_TYPE_USER_START			0x1B2

#define SEQUENCE_EXTENSION_ID			1
#define SEQUENCE_DISPLAY_EXTENSION_ID	2
#define PICTURE_CODING_EXTENSION_ID		8


//////////////////////////////////////////////////////////////////////////
//
//							DataVector
//
//////////////////////////////////////////////////////////////////////////

class DataVector {

private:
	struct DataVectorBlock {
		enum { BLOCK_SIZE = 4096-12 };

		DataVectorBlock *prev;
		DataVectorBlock *next;
		int size;
		char heap[BLOCK_SIZE];
	};

	DataVectorBlock *first, *last;
	const int _item_size;
	unsigned long count;

public:
	DataVector(const int item_size);
	~DataVector();

	void Add(const void *pp);
	void *MakeArray();

	unsigned long Length() { return count; };
};

DataVector::DataVector(int item_size)
	: _item_size(item_size)
{
	first = last = NULL;
	count = 0;
}

DataVector::~DataVector() {
	DataVectorBlock *i, *j;

	i = last;
	while (j = i) {
		i = j->prev;
		delete j;
	}
}

void DataVector::Add(const void *pp) {

	int len = _item_size;

	if (last == NULL) {
		DataVectorBlock *ib = new DataVectorBlock;

		memcpy(ib->heap, pp, len);
		ib->next = NULL;
		ib->size = len;

		if (first == NULL) {
			// New "first" block
			first = ib;
			first->prev = NULL;
		} else {
			// New "last" block
			last = ib;
			first->next = last;
			last->prev = first;
		}

	} else {
		const int remains = DataVectorBlock::BLOCK_SIZE - last->size;

		if (len > remains) {
			// New "last" block
			DataVectorBlock *ib = new DataVectorBlock;

			last->next = ib;
			ib->prev = last;
			ib->next = NULL;
			ib->size = 0;

			if (remains > 0) {
				memcpy(last->heap + last->size, pp, remains);
				pp = (char *)pp + remains;
				last->size += remains;
				len -= remains;
			}

			last = ib;
		}

		memcpy(last->heap + last->size, pp, len);
		last->size += len;
	}

	++count;
}

void *DataVector::MakeArray() {

	char *array = new char[count * _item_size];
	char *ptr = array;

	const DataVectorBlock *dvb = first;

	while (dvb) {
		memcpy(ptr, dvb->heap, dvb->size);
		ptr += dvb->size;

		dvb = dvb->next;
	}

	return array;
}


//////////////////////////////////////////////////////////////////////////
//
//
//					MPEGAudioParser
//
//
//////////////////////////////////////////////////////////////////////////

class MPEGAudioParser {
private:
	unsigned long lFirstHeader;

protected:
	unsigned long header;
	MPEGSampleInfo msi;
	int hstate, skip;

public:
	MPEGAudioParser();

	DataVector audio_stream_blocks;
	DataVector audio_stream_samples;

	virtual void Parse(const unsigned char *, int, unsigned long);
	virtual void Finish() { return; };
};

MPEGAudioParser::MPEGAudioParser()
	: audio_stream_blocks(sizeof(MPEGPacketInfo))
	, audio_stream_samples(sizeof(MPEGSampleInfo))
{
	lFirstHeader = 0;
	header = 0;
	hstate = 0;
	skip = 0;
}

void MPEGAudioParser::Parse(const unsigned char *pData, int len, unsigned long curpkt) {
	const unsigned char *src = pData;

	while (len > 0) {
		if (skip) {
			int tc = skip;

			if (tc > len) tc = len;

			len -= tc;
			skip -= tc;
			src += tc;

			// Audio frame finished?

			if (!skip) {
				msi.pktindex = curpkt;
				msi.pktoffset = src - pData;
				audio_stream_samples.Add(&msi);
			}

			continue;
		}

		// Collect header bytes.

		++hstate;
		header = (header << 8) + *src++;
		--len;

		MPEGAudioHeader hdr(header);

		if (hstate >= 4 && hdr.IsValid()) {

			if (lFirstHeader && !hdr.IsConsistent(lFirstHeader)) {
				continue;
			}

			// Okay, we like the header.

			hstate = 0;

			if (!lFirstHeader) lFirstHeader = header;

			// Setup the sample information.  Don't add the sample, in case it's incomplete.

			msi.header = mpeg_compact_header(header);
			msi.size = hdr.GetFrameSize();

			// Now skip the remainder of the sample.

			skip = msi.size - 4;
		}
	}
}

//////////////////////////////////////////////////////////////////////////
//
//
//					AC3AudioParser
//
//
//////////////////////////////////////////////////////////////////////////

class AC3AudioParser : public MPEGAudioParser {
private:
	unsigned short syncword;
public:
	AC3AudioParser();
	void Parse(const unsigned char *pData, int len, unsigned long curpkt);
};

AC3AudioParser::AC3AudioParser() {
	syncword = 0;
	header = 0;
	hstate = 0;
	skip = 0;
}

void AC3AudioParser::Parse(const unsigned char *pData, int len, unsigned long curpkt) {
	const unsigned char *src = pData;

	while (len > 0) {

		if (skip) {
			int tc = skip;
			if (tc > len) tc = len;

			len -= tc;
			skip -= tc;
			src += tc;

			// Audio frame finished?
			if (!skip) {
				msi.pktindex = curpkt;
				msi.pktoffset = src - pData;
				audio_stream_samples.Add(&msi);
			}
			continue;
		}

		if (syncword == 0x0B77) {
			// Collect header bytes.
			header = (header << 8) + *src++;
			++hstate;
			--len;

			if (hstate == 6) {
				// We have all 8 bytes (include the syncword)
				hstate = 0;

				if (skip = ac3_sync_info(header, NULL, NULL, NULL)) {
					msi.header = ac3_compact_header(header);
					msi.size = skip;
					skip -= 8;
				}

				syncword = 0;
			}
			continue;
		}

		// Search for syncword
		syncword = (syncword << 8) + *src++;
		--len;
	}
}

//////////////////////////////////////////////////////////////////////////
//
//
//					PCMAudioParser
//
//
//////////////////////////////////////////////////////////////////////////


class PCMAudioParser : public MPEGAudioParser {
public:
	PCMAudioParser();
	void Parse(const unsigned char *src, int len, unsigned long curpkt);
	void Finish();
};

PCMAudioParser::PCMAudioParser() {
	header = 0;
	skip = 0;
}

void PCMAudioParser::Parse(const unsigned char *src, int len, unsigned long curpkt) {

	// Since LPCM is raw audio, there isn't anything to parse,
	// thus the "src" data pointer is always NULL except when
	// sending the initial LPCM header (6 bytes).

	if (src != NULL) {

	// src[4] = 5th byte of 6 byte LPCM header:
	// xx......  00=16-,01=20-,10=24-bit (11=forbidden)
	// ..xx....  00=48,01=96 KHz (others reserved)
	// ....x...  reserved
	// .....xxx  number of channels - 1

		const int chans = (src[4] & 7) + 1;
		const int bits = 16 + ((src[4] & 0xC0) >> 4);

		skip = chans * (PCM_BUFFER_SIZE / 8) * bits;

		msi.header = src[4];	// this never changes
		msi.pktindex = 0;
		msi.pktoffset = 0;
		msi.size = 0;
		return;
	}

	msi.size += len;

	while (msi.size >= skip) {
		int tc = msi.size - skip;

		// Finish off this block
		msi.size = skip;
		audio_stream_samples.Add(&msi);

		// Start a new block
		msi.size = tc;
		msi.pktindex = curpkt;
		msi.pktoffset = len - tc;
	}
}

void PCMAudioParser::Finish() {
	// Finish off the final block

	if (msi.size > 0) {
		// Make sure we keep only whole samples
		const int chans = (msi.header & 7) + 1;
		const int bits = 16 + ((msi.header & 0xC0) >> 4);

		msi.size -= msi.size % (chans * bits);
		if (msi.size > 0) {
			audio_stream_samples.Add(&msi);
		}
	}
}


//////////////////////////////////////////////////////////////////////////
//
//
//					MPEGVideoParser
//
//
//////////////////////////////////////////////////////////////////////////


class MPEGVideoParser {
private:
	unsigned char buf[8];

	int idx, bytes;

	MPEGSampleInfo msi;
	__int64 bytepos;
	__int64 last_access_unit;

	long header;

	bool fPicturePending;
	bool fFoundSequenceStart;

	bool sequence_header();
	bool picture_start_code();
	void extension_start_code();

public:

	VDXFraction mFrameRate;
	DataVector video_stream_samples;

	int width, height;
	long seq_ext;
	int aspect_ratio_info;
	bool progressive_sequence;
//	bool fIgnoreSeqHdrs;

	MPEGVideoParser();

	bool Parse(const unsigned char *pData, int len, unsigned long curpkt);
};

MPEGVideoParser::MPEGVideoParser()
	: video_stream_samples(sizeof(MPEGSampleInfo))
{
	bytepos = 0;
	header = -1;
//	fIgnoreSeqHdrs = false;

	fPicturePending = false;
	fFoundSequenceStart = false;
	seq_ext = 0;

	last_access_unit = 0;

	idx = bytes = 0;
	msi.frame_type = 0;
}

inline bool MPEGVideoParser::picture_start_code() {

// xxxxxxxx xx......	temporal reference
// ........ ..xxx...	picture coding type
// ........ ........

	header = 0xFFFFFFFF;

	// Fix for Shakira: don't add picture if it is corrupt

	int ft = (buf[1] >> 3) & 7;

	switch (ft) {
		case MPEG_FRAME_TYPE_I:
		case MPEG_FRAME_TYPE_P:
		case MPEG_FRAME_TYPE_B:
			break;
		default:
			return false;
	}

	msi.frame_type = (char)ft;
	msi.bitflags = PICTUREFLAG_PF | FRAME_PICTURE;
	fPicturePending = true;

	return true;
}


// Fix for Kylie: limit what res we will support
#define MPEG_MAX_WIDTH		1920
#define MPEG_MAX_HEIGHT		1088
// 1920*1080*4 = 8MB buffer

inline bool MPEGVideoParser::sequence_header() {

	static const VDXFraction frates[16] = {
		{15000, 1000},		// 0000 forbidden
		{24000, 1001},		// 0001 23.976
		{24000, 1000},		// 0010 24
		{25000, 1000},		// 0011 25
		{30000, 1001},		// 0100 29.97
		{30000, 1000},		// 0101 30
		{50000, 1000},		// 0110 50
		{60000, 1001},		// 0111 59.94
		{60000, 1000},		// 1000 60
		{15000, 1000},		// 1001 reserved
		{15000, 1000},		// 1010 reserved
		{15000, 1000},		// 1011 reserved
		{15000, 1000},		// 1100 reserved
		{15000, 1000},		// 1101 reserved
		{15000, 1000},		// 1110 reserved
		{15000, 1000}		// 1111 reserved
	};

// bytes 0 to 3
// xxxxxxxx xxxx.... ........ ........	vertical_size
// ........ ....xxxx xxxxxxxx ........	horizontal_size
// ........ ........ ........ xxxx....	pel_aspect_ratio
// ........ ........ ........ ....xxxx	picture_rate

// bytes 4 to 7
// xxxxxxxx xxxxxxxx xx...... ........  bit_rate
// ........ ........ ..1..... ........  marker_bit
// ........ ........ ...xxxxx xxxxx...  vbv_buffer_size
// ........ ........ ........ .....x..  constrained_parameters_flag
// ........ ........ ........ ......x.  load_intra_quantiser_matrix
// ........ ........ ........ .......x  load_non_intra_quantiser_matrix

	header = 0xFFFFFFFF;

	// Verify that this is a valid sequence header
	const int w = (buf[0] << 4) | (buf[1] >> 4);
	const int h = ((buf[1] & 15) << 8) | buf[2];
	const int ar = buf[3] >> 4;
	const int pr = buf[3] & 15;
	const int br = (buf[4] << 10) | (buf[5] << 2) | (buf[6] >> 6);

	// For robustness, we will only fail if:
	//	a) This appears to be a valid sequence header, and
	//	b) width or height doesn't match previous sequence headers

	if (w > 0 && w <= MPEG_MAX_WIDTH
		&& h > 0 && h <= MPEG_MAX_HEIGHT
		&& ar != 0 && pr != 0 && br != 0
		&& (buf[6] & 0x20) != 0) {

		if (fFoundSequenceStart) {
			// Check this against previous sequence headers
			// For robustness, only check width and height

//			if (!fIgnoreSeqHdrs) {
				if (w != width || h != height) return false;
//			}

		} else {
			width = w;
			height = h;
			aspect_ratio_info = ar;
			mFrameRate = frates[pr];

			fFoundSequenceStart = true;
		}
	}

	return true;
}

inline void MPEGVideoParser::extension_start_code() {

	// EXTENSION_START_CODE

	if (bytes == 1) {
		int ext = buf[0] >> 4;

		if (ext == SEQUENCE_EXTENSION_ID
			&& fFoundSequenceStart
			&& seq_ext == 0) {

			bytes = 4;
			return;
		}

		if (ext == PICTURE_CODING_EXTENSION_ID
			&& fPicturePending) {

			bytes = 5;
			return;
		}

	} else if (bytes == 4) {

		seq_ext = *(long *)buf;

// xxxx.... ........ ........	SEQUENCE_EXTENSION_ID
// ....x... ........ ........	escape_bit
// .....xxx xxxx.... ........	MPEG-2 profile / level
// ........ ....x... ........	progressive_sequence
// ........ .....xx. ........	chroma_format
// ........ .......x x.......	horizontal_size_extension
// ........ ........ .xx.....	vertical_size_extension

		progressive_sequence = (buf[1] >> 3) & 1;
		width = (width & 0xFFF) | ((buf[1] & 1) << 13) | ((buf[2] & 0x80) << 5);
		height = (height & 0xFFF) | ((buf[2] & 0x60) << 7);

	} else if (bytes == 5) {

// xxxx.... ........ ........ ........ ........	PICTURE_CODING_EXTENSION_ID
// ....xxxx xxxxxxxx xxxx.... ........ ........	f_codes
// ........ ........ ....xx.. ........ ........	intra_dc_precision
// ........ ........ ......xx ........ ........	picture_structure
// ........ ........ ........ x....... ........	top_field_first
// ........ ........ ........ .x...... ........	frame_pred_frame_dct
// ........ ........ ........ ..x..... ........	concealment_vectors
// ........ ........ ........ ...x.... ........	q_scale_type
// ........ ........ ........ ....x... ........	intra_vlc_format
// ........ ........ ........ .....x.. ........	alternate_scan
// ........ ........ ........ ......x. ........	repeat_first_field
// ........ ........ ........ .......x ........	chroma_420_type
// ........ ........ ........ ........ x.......	progressive_frame

		msi.bitflags = buf[2] & 3;
		if (buf[3] & 0x80) msi.bitflags |= PICTUREFLAG_TFF;
		if (buf[3] & 0x02) msi.bitflags |= PICTUREFLAG_RFF;
		if (buf[4] & 0x80) msi.bitflags |= PICTUREFLAG_PF;
	}

	header = 0xFFFFFFFF;
}

bool MPEGVideoParser::Parse(const unsigned char *pData, int len, unsigned long curpkt) {

	const unsigned char *src = pData;

	while (len > 0) {

		// bytes is zero when the class is created,
		// thus (idx < bytes) is false the first time

		if (idx < bytes) {
			int tc = bytes - idx;
			if (tc > len) tc = len;

			memcpy(buf + idx, src, tc);

			len -= tc;
			idx += tc;
			src += tc;

			if (idx >= bytes) {   // Finished?

				switch (header) {

				case VIDPKT_TYPE_PICTURE_START:
					picture_start_code();
					break;

				case VIDPKT_TYPE_SEQUENCE_START:
					if (!sequence_header()) return false;

					// A sequence header is ALWAYS an access unit
					last_access_unit = bytepos + ((src - 12) - pData);
					break;

				case VIDPKT_TYPE_EXT_START:
					extension_start_code();
					break;
				}
			}
			continue;
		}

		// Look for a valid MPEG header
		header = (header << 8) + *src++;
		--len;

		if ((header & 0xFFFFFF00) == 0x100) {

			if (fPicturePending
				&& (header == VIDPKT_TYPE_SEQUENCE_START
				|| header == VIDPKT_TYPE_GROUP_START
				|| header == VIDPKT_TYPE_PICTURE_START
				|| header == VIDPKT_TYPE_SEQUENCE_END))
			{

				__int64 pos = last_access_unit;

				msi.pktindex = curpkt;
				msi.pktoffset = src - pData;

				// The spec defines all of the above as
				// access units, except for sequence end,
				// which is part of the current access unit.


				last_access_unit = bytepos + ((src - 4) - pData);

				// (sequence_end is not useful for our purpose here)
//				if (header == VIDPKT_TYPE_SEQUENCE_END) last_access_unit += 4;

				msi.size = (long)(last_access_unit - pos);
				video_stream_samples.Add(&msi);

				fPicturePending = false;
			}

			// (a - b) - c  =  (a - c) - b

			switch (header) {

			case VIDPKT_TYPE_SEQUENCE_START:
				idx = 0;
				bytes = 8;
				break;

			case VIDPKT_TYPE_PICTURE_START:
				// Don't add pictures until we've parsed the sequence!
				if (fFoundSequenceStart) {
					idx = 0;
					bytes = 2;
				}
				break;

			case VIDPKT_TYPE_EXT_START:
				if (fFoundSequenceStart) {
					idx = 0;
					bytes = 1;
				}
				break;

//			case VIDPKT_TYPE_SEQUENCE_END:
//				if (fFoundSequenceStart) {
//					if (!fIgnoreSeqHdrs) return false;
//				}
			
			default:
				header = 0xFFFFFFFF;
				break;
			}
		}
	}

	bytepos += src - pData;
	return true;
}


//////////////////////////////////////////////////////////////////////////
//
//
//							MPEGFileParser
//
//
//////////////////////////////////////////////////////////////////////////

#define NON_INTERLEAVED_PACK_SIZE	16384

class MPEGFileParser : public IMPEGFileParser {
private:
	InputFileMPEG2 *const parentPtr;

	// Fix for M2M, search at least 3 MB
	enum {
		SEARCH_SYSTEM_LIMIT = 1048576*3,
		SEARCH_AUDIO_LIMIT  = 256
	};

	DataVector video_stream_blocks;

	unsigned char packet_buffer[65536];

	inFile *mFile;
	int vidstream;
	bool fAbort;
	HWND hWndStatus;

	int NextStartCode();
	bool Skip(int bytes);

	int Read();
	int Read(void *, int);

	bool Verify();
	__int64 PacketPTS(int pkt, int stream_id);
	void DoFixUps();

	bool ParsePack(int stream_id);
	bool guiDlgMessageLoop(HWND hDlg);

	static INT_PTR CALLBACK ParseDialogProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);

	MPEGVideoParser *videoParser;
	MPEGAudioParser *audioParser[48];

	bool fIsMPEG2;
	bool fInterleaved;
	bool fIsScrambled;
	long lFirstCode;

	int audio_streams;

public:
	MPEGFileParser(InputFileMPEG2 *const pp);
	void Destruct();
//	~MPEGFileParser();

	void Parse();
};

MPEGFileParser::MPEGFileParser(InputFileMPEG2 *const pp)
	: parentPtr(pp)
	, video_stream_blocks(sizeof(MPEGPacketInfo))
{
	_RPT0(_CRT_WARN, "MPEGFileParser constructed\n");

	// Clear variables

	videoParser = NULL;
	vidstream = 0;

	audio_streams = 0;
	for (int i = 0; i < 48; ++i) {
		audioParser[i] = NULL;
	}

	fIsScrambled = false;
	fAbort = false;

	hWndStatus = NULL;
	mFile = &(parentPtr->mFile);
}

//MPEGFileParser::~MPEGFileParser() {
void MPEGFileParser::Destruct() {
	_RPT0(_CRT_WARN, "MPEGFileParser destructed\n");

	if (hWndStatus) DestroyWindow(hWndStatus);

	for (int i = 0; i < 48; ++i) {
		delete audioParser[i];
	}

	delete videoParser;
	delete this;
}

int MPEGFileParser::NextStartCode() {
	int i;

/*
	while (!mFile->inEOF()) {
		i = mFile->inByte();
		if (i == 0) {
			i = mFile->inByte();
			while (i == 0) {
				i = mFile->inByte();
				if (mFile->inEOF()) break;
				if (i == 1) {
					return mFile->inByte() + 0x100;
				}
			}
		}
	}
*/

	i = 0xFFFFFFFF;
	while (!mFile->inEOF()) {
		i = (i << 8) + mFile->inByte();
		if ((i & 0xFFFFFF00) == 0x100) return i;
	}

	return 0;
}

bool MPEGFileParser::Skip(int bytes) {
	return mFile->inSeek(mFile->inPos() + bytes);
}

int MPEGFileParser::Read() {
	if (!mFile->inEOF()) return mFile->inByte();
	return EOF;
}

int MPEGFileParser::Read(void *buffer, int bytes) {
	return mFile->inRead(buffer, bytes);
}

inline bool MPEGFileParser::Verify() {

	// Check the validity of an MPEG file

	unsigned char buf[4];
	unsigned long hdr;
	long lFirstPES;
	long lFirstSeq;
    long count;
	int i;

	fIsMPEG2 = false;
	fInterleaved = false;
	lFirstCode = -1;
	lFirstPES = -1;
	lFirstSeq = -1;

	// Search for system layer
	// Emotion.m2v has empty PES headers (0x1E0) but no system!

	mFile->inSeek(0);
	Read(buf, 4);
	hdr = (buf[0] << 24) + (buf[1] << 16) + (buf[2] << 8) + buf[3];
	count = 4;

	while (count < SEARCH_SYSTEM_LIMIT) {

		if (hdr == 0x1BA) {

			// Verify pack header
			// (This is very robust)

			Read(buf, 4);
			count += 4;

			if (	(buf[0] & 0xC4) == 0x44
				&&	(buf[2] & 4) == 4) {

			// MPEG-2 pack
			// 01sss1ss ssssssss sssss1ss ssssssss sssss1ee
			// eeeeeee1 rrrrrrrr rrrrrrrr rrrrrr11 11111xxx

			// If we fail, buf[3] might still begin
			// a start code, so don't toss it yet!

				buf[0] = buf[3];

				Read(buf+1, 3);
				count += 3;

				if (	(buf[1] & 4) == 4
					&&	(buf[2] & 1) == 1) {

					// Almost there!

					buf[0] = buf[3];

					Read(buf+1, 3);
					count += 3;

					const long mux	= (buf[0] << 14)
									+ (buf[1] << 6)
									+ (buf[2] >> 2);

					if (mux && (buf[2] & 3) == 3) {
						// Yes!!

						Skip(buf[3] & 7);
						fInterleaved = true;
						fIsMPEG2 = true;
						lFirstCode = count - 14;
						break;
					}
				}

			} else if (	(buf[0] & 0xF1) == 0x21
					&&	(buf[2] & 1) == 1) {

			// MPEG-1 pack
			// 0010aaa1 aaaaaaaa aaaaaaa1 aaaaaaaa
			// aaaaaaa1 1mmmmmmm mmmmmmmm mmmmmmm1

				buf[0] = buf[3];

				Read(buf+1, 3);
				count += 3;

				if (	(buf[1] & 1) == 1
					&&	(buf[2] & 0x80) == 0x80) {

					// Almost there!

					buf[0] = buf[1];
					buf[1] = buf[2];
					buf[2] = buf[3];
					buf[3] = (unsigned char)Read();
					++count;

					const long mux	= (buf[1] << 15)
									+ (buf[2] << 7)
									+ (buf[3] >> 1);

					if (mux && (buf[3] & 1) == 1) {
						// Yes!!

						fInterleaved = true;
						lFirstCode = count - 12;
						break;
					}
				}
			}

			// Damn, we were fooled
			// Back up four bytes

			hdr = (buf[0] << 24) + (buf[1] << 16) + (buf[2] << 8) + buf[3];
			count -= 4;
			continue;
		}

		// .vdr files contain PES packets, but no pack headers!
		// To support these, let's say that we must find at least
		// two PES packets AND one sequence start code...

		if ((hdr >= 0x1C0 && hdr <= 0x1EF) || hdr == 0x1BD) {

			Read(buf, 4);
			count += 4;

			// The legal values for buf[2] are:
			// 10.. ....  MPEG-2
			// 01.. ....  MPEG-1
			// 0010 ....  MPEG-1
			// 0011 ....  MPEG-1
			// 0000 1111  MPEG-1
			// 1111 1111  MPEG-1

			if ((buf[0] != 0 || buf[1] != 0) &&
				(buf[2] == 0x0F || buf[2] == 0xFF ||
				(buf[2] & 0xC0) == 0x80 ||
				(buf[2] & 0xC0) == 0x40 ||
				(buf[2] & 0xE0) == 0x20)) {

				if (lFirstPES != -1) {
					if (lFirstSeq != -1) {
						fIsMPEG2 = ((buf[2] & 0xC0) == 0x80);
						lFirstCode = lFirstPES;
						fInterleaved = true;
						break;
					}
				} else {
					lFirstPES = count - 8;
				}
			}

			// Fooled; back up four bytes
			hdr = (buf[0] << 24) + (buf[1] << 16) + (buf[2] << 8) + buf[3];
			count -= 4;
			continue;
		}

		// If we hit a sequence header, remember it

		if (hdr == VIDPKT_TYPE_SEQUENCE_START) {

			Read(buf, 4);
			count += 4;

			// wwwwwwww wwwwhhhh hhhhhhhh aaaarrrr
			// bbbbbbbb bbbbbbbb bb1vvvvv vvvvvxx.

			const int w = (buf[0] << 4) + (buf[1] >> 4);
			const int h = ((buf[1] & 15) << 4) + buf[2];
			const int ar = buf[3] >> 4;
			const int pr = buf[3] & 15;

			if (w > 0 && w <= MPEG_MAX_WIDTH
				&& h > 0 && h <= MPEG_MAX_HEIGHT
				&& ar != 0 && pr != 0) {

				// In this case, we can discard the whole buffer
				// because no start code can begin at any byte

				Read(buf, 4);
				count += 4;

				const int br = (buf[0] << 10) + (buf[1] << 2) + (buf[2] >> 6);

				if (br != 0 && (buf[2] & 0x20) == 0x20) {
					// Yay!

					// Let's say now that IF we've found multiple
					// valid sequence headers, AND we've searched
					// ~70000 bytes for system headers, then it's
					// fairly certain this is an elementary stream

					if (lFirstSeq != -1) {
						// We found at least one sequence
						// header prior to this one
						if (count >= 70000) {
							lFirstCode = lFirstSeq;
							break;
						}
					} else {
						lFirstSeq = count - 12;
					}
				}
			}

			// Damn, fooled again
			// Back up four bytes

			hdr = (buf[0] << 24) + (buf[1] << 16) + (buf[2] << 8) + buf[3];
			count -= 4;
			continue;
		}

		if ((i = Read()) == EOF) break;
		hdr = (hdr << 8) + i;
		++count;
    }

    if (lFirstCode < 0) {
		// If we found at least one sequence header, use that
		if (lFirstSeq != -1) {
			lFirstCode = lFirstSeq;
		} else {
			return false;
		}
	}

	if (!fInterleaved) {
		// No system layer...
		// Jump back to the sequence header,
		// and check for sequence extension

		mFile->inSeek(lFirstCode + 8);

		if (NextStartCode() == VIDPKT_TYPE_EXT_START) {
			fIsMPEG2 = ((Read() >> 4) == SEQUENCE_EXTENSION_ID);
		}
	}

	// Looks good to me
	return true;
}

bool MPEGFileParser::ParsePack(int stream_id) {
	// Parse one elementary stream packet header
	// Return "false" if packet is incomplete
	MPEGPacketInfo mpi;
	long pack_length;
	int i;

	bool complete = false;

	do {

		if (fInterleaved) {
			pack_length = Read();
			pack_length = (pack_length << 8) + Read();

			// Pack length of zero is not an error
			if (pack_length == 0) {
				complete = true;
				break;
			}

			if ((stream_id >= 0x1C0 && stream_id <= 0x1EF) || stream_id == 0x1BD) {
				// MPEG Audio, MPEG Video, or private_stream_1

				// Parse PES header
				i = Read();
				--pack_length;

				if ((i & 0xC0) == 0x80) {	// MPEG-2 PES
					int data_length;

					if (i & 0x30) fIsScrambled = true;

					i = Read();				// PTS/DTS flags
					data_length = Read();	// PES_header_data_length
					pack_length -= 2;

					if (i & 0x80) {			// PTS follows
						Skip(5);
						pack_length -= 5;
						data_length -= 5;
						if (i & 0x40) {		// DTS follows
							Skip(5);
							pack_length -= 5;
							data_length -= 5;
						}
					}

					if (data_length > 0) {
						Skip(data_length);
						pack_length -= data_length;
					}

				} else {				// MPEG-1 PES

					while (i & 0x80) {
						if (--pack_length <= 0) break;
						i = Read();
					}

					if (i & 0x40) {
						Skip(1);		// P-STD_buffer size
						i = Read();
						pack_length -= 2;
					}

					if (i & 0x20) {		// PTS follows
						Skip(4);
						pack_length -= 4;
						if (i & 0x10) {	// DTS follows
							Skip(5);
							pack_length -= 5;
						}
					}
				}
            }

			if (pack_length <= 0) {
				// PES packet with no payload is not an error
				complete = (pack_length == 0);
				break;
			}

		} else {
			// Elementary video
			stream_id = 0x1E0;
			pack_length = NON_INTERLEAVED_PACK_SIZE;
		}
        
		if (stream_id == 0x1BD) {
			// private_stream_1, AC3 or LPCM
			
			i = Read();		// substream
			--pack_length;
			
			if ((i & 0xF8) == 0x80) {
				// Dolby Digital AC-3, substreams 0x80 to 0x87
				int id = parentPtr->audio_stream_list[i - 128];

				if (id == 0) {
					// map to parsers 32 to 39
					if (audioParser[i - 96] = new AC3AudioParser) {
						id = i - 95;
						parentPtr->audio_stream_list[i - 128] = id;
					}
				}

				if (id > 0) {
					--id;

			// 1 byte  = # of syncwords (i.e., access units) in packet
			// 2 bytes = 1-based offset of first syncword (0x0B77)
					Skip(3);
					pack_length -= 3;

					mpi.file_pos = (mFile->inPos() << 16) + pack_length;
					audioParser[id]->audio_stream_blocks.Add(&mpi);
					//audio_stream_blocks[id]->Add(&mpi);

					i = Read(packet_buffer, pack_length);
					complete = (i >= pack_length);
					audioParser[id]->Parse(packet_buffer, i,
						audioParser[id]->audio_stream_blocks.Length() - 1);
					pack_length = 0;
				}
			
			} else if ((i & 0xF8) == 0xA0) {
				// Uncompressed LPCM

				unsigned char buf[6];
				int frame_offset;

				Read(buf, 6);
				pack_length -= 6;

				frame_offset = (buf[1] << 8) + buf[2];

	//	0:	xxxxxxxx	number of frames
	//	1:	xxxxxxxx	offset of first frame (MSB)
	//	2:	xxxxxxxx	offset of first frame (LSB)
	//	3:	x.......	emphasis
	//		.x......	mute
	//		..x.....	reserved
	//		...xxxxx	frame number % 20
	//	4:	xx......	00=16-,01=20-,10=24-bit (11=forbidden)
	//		..xx....	00=48,01=96 KHz (others reserved?)
	//		....x...	reserved
	//		.....xxx	number of channels - 1
	//	5:	xxxxxxxx	dynamic range control

	// We can support 48K or 96K 16- or 24-bit mono or stereo

				if (	buf[0] > 0
					&&	frame_offset > 0
					&&	frame_offset < pack_length
					&&	(buf[4] & 0xC0) != 0xC0
					&&	(buf[4] & 0x26) == 0) {

					// Looks good
					// Uncompressed LPCM 0xA0 to 0xA7
					int id = parentPtr->audio_stream_list[i - 128];

					if (id == 0) {
						// map to parsers 40 to 47
						if (audioParser[i - 120] = new PCMAudioParser) {
							id = i - 119;
							parentPtr->audio_stream_list[i - 128] = id;

							// Prime him with the first header
							audioParser[i - 120]->Parse(buf, 6, 0);
						}
					}

					if (id > 0) {
						--id;

						mpi.file_pos = (mFile->inPos() << 16) + pack_length;
						//audio_stream_blocks.Add(&mpi);
						audioParser[id]->audio_stream_blocks.Add(&mpi);

						i = Read(packet_buffer, pack_length);
						complete = (i >= pack_length);

						audioParser[id]->Parse(NULL, i,
							audioParser[id]->audio_stream_blocks.Length() - 1);
						pack_length = 0;
					}
				}
			}

		} else if (stream_id >= 0x1E0 && stream_id <= 0x1EF) {

			// MPEG video (only the first stream found)
			if (!vidstream) vidstream = stream_id;

			if (stream_id == vidstream) {
				mpi.file_pos = (mFile->inPos() << 16) + pack_length;
				video_stream_blocks.Add(&mpi);
				i = Read(packet_buffer, pack_length);
				complete = (i >= pack_length);
				if (!videoParser->Parse(packet_buffer, i, video_stream_blocks.Length() - 1)) {
					complete = false;	// end of sequence
				}
				pack_length = 0;
			}

		} else if (stream_id >= 0x1C0 && stream_id <= 0x1DF) {
			// MPEG audio 0xC0 to 0xDF

			int id = parentPtr->audio_stream_list[stream_id - 0x180];

			if (id == 0) {
				// map to parsers 0 to 31
				if (audioParser[stream_id - 0x1C0] = new MPEGAudioParser) {
					id = stream_id - 0x1BF;
					parentPtr->audio_stream_list[stream_id - 0x180] = id;
				}
			}

			if (id > 0) {
				--id;

				mpi.file_pos = (mFile->inPos() << 16) + pack_length;
				//audio_stream_blocks.Add(&mpi);
				audioParser[id]->audio_stream_blocks.Add(&mpi);

				i = Read(packet_buffer, pack_length);
				complete = (i >= pack_length);
				audioParser[id]->Parse(packet_buffer, i,
					audioParser[id]->audio_stream_blocks.Length() - 1);
				pack_length = 0;
			}
		}

    } while (0);

	if (pack_length > 0) return Skip(pack_length);
	return complete;
}

__int64 MPEGFileParser::PacketPTS(int pkt, int stream_id) {
	unsigned long hdr;

	// Return the Presentation Time Stamp for PES
	// packet "pkt" of stream number "stream_id"

	// For video packets, stream_id >= 0x1E0 and <= 0x1EF
	// For audio packets, it is the stream number (0 to 47)

	__int64 offset;
	__int64 limit;
	int id;

	stream_id &= 0xFF;

	if (stream_id < 0xE0) {
		// Audio
		id = stream_id;

		// For private_stream_1, we need to account
		// for the following:
		// substream number is 1 byte
		// AC-3 has an additional 3 byte header
		// LPCM has an additional 6 byte header

		if (id < 32) {				// MPEG
			stream_id += 0x1C0;
			limit = 0;
		} else if (id < 40) {		// AC-3
			stream_id = 0x1BD;
			limit = -4;
		} else {					// LPCM
			stream_id = 0x1BD;
			limit = -7;
		}

		limit += parentPtr->audio_packet_list[id][pkt].file_pos >> 16;

	} else {

		// Video
		limit = parentPtr->video_packet_list[pkt].file_pos >> 16;
		stream_id += 0x100;
	}

	// The largest possible PES packet header is 264
	// bytes (cf. ISO/IEC 13818-1), therefore we start
	// searching at least that far behind the PES data:

	// bytes      description
	//  3         packet_start_code_prefix
	//  1         stream_id
	//  2         PES_packet_length
	//  2         miscellaneous flags
	//  1         PES_header_data_length
	//  0..255	  remainder of PES header

	// After parsing each PES header, we'll know we have
	// the right PES packet if our position exactly matches
	// the sample's starting position.

	offset = limit - 264;
	if (offset < 0) offset = 0;

	mFile->inSeek(offset);
	hdr = 0xFFFFFFFF;

	while (mFile->inPos() < limit) {

		// Just blaze through the bytes
		hdr = (hdr << 8) + Read();

		if (hdr == (unsigned long)stream_id) {
			// This may be the packet we want...
			// Parse the PTS (if any) in the PES header
			__int64 pts;

			Skip(2);	// pack length
			int ptsflags = Read();

			if ((ptsflags & 0xC0) == 0x80) {	// MPEG-2
				int data_length;

				ptsflags = Read();
				data_length = Read();

				if (ptsflags & 0x80) {
			// 0010ppp1 pppppppp ppppppp1 pppppppp ppppppp1
					pts = (Read() & 0x0E) >> 1;
					pts = (pts << 8) + Read();
					pts = (pts << 7) + (Read() >> 1);
					pts = (pts << 8) + Read();
					pts = (pts << 7) + (Read() >> 1);
					data_length -= 5;
				}
				Skip(data_length);

			} else {
				// MPEG-1
				while (ptsflags & 0x80) {
					ptsflags = Read();
					// Don't endless loop
					if (mFile->inPos() >= limit) break;
				}
				if (ptsflags & 0x40) {
					Read();
					ptsflags = Read();
				}
				if (ptsflags & 0x20) {
					pts = (ptsflags & 0x0E) >> 1;
					pts = (pts << 8) + Read();
					pts = (pts << 7) + (Read() >> 1);
					pts = (pts << 8) + Read();
					pts = (pts << 7) + (Read() >> 1);
					if (ptsflags & 0x10) Skip(5);
					ptsflags = 0x80;
				}
			}

			// Finally, if this is the PES packet we want,
			// our position will exactly match the stored
			// position of the PES data for the sample:

			if (mFile->inPos() == limit) {
				if (ptsflags & 0x80) return pts;
				break;	// No PTS on this packet
			}
		}
	}

	return -1;	// PTS not found
}

inline void MPEGFileParser::DoFixUps() {
	MPEGSampleInfo *msi;
	unsigned long index;
	long offset;
	int i, limit;
	int id;

	// Determine the packets in which each video and audio
	// sample originates, and the offsets in those packets.

	// The parser has stored the packet and offset at which
	// the END of the sample was found. (In the case of video,
	// it's actually 4 bytes past the end of the sample.)

	for (i = 0; i < parentPtr->vframes; ++i) {
		msi = &(parentPtr->video_sample_list[i]);
		index = msi->pktindex;
		offset = msi->pktoffset - (msi->size + 4);

		if (offset < 0) {
			// Sample began in an earlier packet
			while (index) {
				--index;
				offset += (long)(parentPtr->video_packet_list[index].file_pos & 0xFFFF);
				if (offset >= 0) {
					msi->pktindex = index;
					msi->pktoffset = (unsigned short)offset;
					break;
				}
			}
		} else {
			// Sample begins in this packet
			msi->pktoffset = (unsigned short)offset;
		}
	}

	// Note: LPCM doesn't need fixing up

	for (id = 0; id < 40; ++id) {
		for (i = 0; i < parentPtr->aframes[id]; ++i) {
			msi = &(parentPtr->audio_sample_list[id][i]);
			index = msi->pktindex;
			offset = msi->pktoffset - msi->size;

			if (offset < 0) {
				// Sample began in an earlier packet
				while (index) {
					--index;
					offset += (long)(parentPtr->audio_packet_list[id][index].file_pos & 0xFFFF);
					if (offset >= 0) {
						msi->pktindex = index;
						msi->pktoffset = (unsigned short)offset;
						break;
					}
				}
			} else {
				// Sample begins in this packet
				msi->pktoffset = (unsigned short)offset;
			}
		}
	}

	// Now hunt for associated Presentation Time Stamps
	// They are required at least every 0.7 seconds;
	// We'll give 'em about 2 seconds

	parentPtr->vfirstPTS = 0;
	for (id = 0; id < 48; ++id) {
		parentPtr->afirstPTS[id] = 0;
	}
	if (!fInterleaved) return;

	// Find the first I-frame with PTS

	limit = (int)((2.0 * (double)videoParser->mFrameRate.mNumerator) /
		(double)videoParser->mFrameRate.mDenominator + 0.5);

	if (limit > parentPtr->vframes) limit = parentPtr->vframes;

	for (i = 0; i < limit; ++i) {
		const MPEGSampleInfo *msi = &(parentPtr->video_sample_list[i]);

		if (msi->frame_type == MPEG_FRAME_TYPE_I) {
			__int64 pts = PacketPTS(msi->pktindex, vidstream);
			if (pts != -1) {
				// Got it!
				// Remove all frames prior to this one
				if (i > 0) {
					for (int j = i; j < parentPtr->vframes; ++j) {
						parentPtr->video_sample_list[j - i] =
						parentPtr->video_sample_list[j];
					}
				}
				parentPtr->vframes -= i;
				parentPtr->vfirstPTS = pts;
				break;
			}
		}
	}

	// Find the first audio frame with PTS

	for (id = 0; id < 48; ++id) {

		if (parentPtr->aframes[id] > 0) {
			long srate;
			long lFramesPerSec;
			unsigned short header = parentPtr->audio_sample_list[id][0].header;

			if (id < 32) {
				MPEGAudioCompacted hdr(header);
				srate = (long)hdr.GetSamplingRateHz();
				if (hdr.GetLayer() == 1) {
					lFramesPerSec = (long)((double)srate / 384.0 + 0.5);
				} else {
					lFramesPerSec = (long)((double)srate / 1152.0 + 0.5);
				}
			} else if (id < 40) {
				if (ac3_compact_info(header, &srate, NULL, NULL)) {
					lFramesPerSec = (long)((double)srate / 1536.0 + 0.5);
				}
			} else {
				srate = (header & 0x30)? 96000: 48000;
				lFramesPerSec = (long)((double)srate / 1536.0 + 0.5);
			}

			limit = lFramesPerSec * 2;
			if (limit > parentPtr->aframes[id]) limit = parentPtr->aframes[id];

			for (i = 0; i < limit; ++i) {
				const MPEGSampleInfo *msi = &(parentPtr->audio_sample_list[id][i]);
				__int64 pts = PacketPTS(msi->pktindex, id);
				if (pts != -1) {
					// Got it!
					// Remove all frames prior to this one
					if (i > 0) {
						for (int j = i; j < parentPtr->aframes[id]; ++j) {
							parentPtr->audio_sample_list[id][j - i] =
								parentPtr->audio_sample_list[id][j];
						}
						parentPtr->aframes[id] -= i;
					}
					parentPtr->afirstPTS[id] = pts;
					break;
				}
			}
		}
	}
}

void MPEGFileParser::Parse() {
	int i;

	// Verify the MPEG file.  If successful,
	// This sets up the following variables:
	//		lFirstCode
	//		fInterleaved
	//		fIsMPEG2

	if (!Verify()) {
		parentPtr->mContext.mpCallbacks->SetError("Couldn't detect file type");
		return;
	}

	if (!fInterleaved && lFirstCode >= 4) {
		// Beware of Transport Streams!
		// These will appear valid, however:
		//	- fInterleaved will be false
		//	- lFirstCode will be at least 4
		//	- The first byte will be 0x47

		mFile->inSeek(0);
		if (Read() == 0x47) {
			MessageBox(NULL, "File matches the profile of a "
				"Transport Stream and may not decode correctly.",
				"Warning", MB_ICONEXCLAMATION | MB_SETFOREGROUND | MB_TASKMODAL | MB_OK);
		}
	}

	// Begin file parsing!  This is a royal bitch!

	videoParser = new MPEGVideoParser;

//	videoParser->fIgnoreSeqHdrs = parentPtr->fIgnoreSeqHdrs;

	mFile->inSeek(lFirstCode);

	// pop up the dialog...
	hWndStatus = CreateDialogParam(g_hModule,
		MAKEINTRESOURCE(IDD_PROGRESS), GetActiveWindow(),
		ParseDialogProc, (LPARAM)this);

	do {
		if (fInterleaved) {
			i = NextStartCode();

			if (i >= 0x1BB) {
				if (!ParsePack(i)) break;

			} else {
				if (i == 0x1B9) break;	// ISO 11172 end code
				if (i == 0) break;		// end of file
			}

		} else {

			// Elementary video
			if (!ParsePack(0x1E0)) break;
		}

		if (hWndStatus != NULL) {
			if (!guiDlgMessageLoop(hWndStatus)) break;
		}

	} while (!fAbort);

	parentPtr->fUserAborted = fAbort;

	// Finish off any final LPCM audio packets
	for (i = 40; i < 48; ++i) {
		if (audioParser[i] != NULL) {
			audioParser[i]->Finish();
		}
	}

    // Construct stream and packet lookup tables, and
	// transfer them to the parent (I don't like this, but...)

	parentPtr->video_packet_list = (MPEGPacketInfo *)
        video_stream_blocks.MakeArray();
    parentPtr->vpackets = video_stream_blocks.Length();

	if (fInterleaved) {
		int id;

		for (i = 0; i < 96; ++i) {
			if (parentPtr->audio_stream_list[i] != 0) {
				id = parentPtr->audio_stream_list[i] - 1;

				if (audioParser[id]) {
					parentPtr->audio_packet_list[id] = (MPEGPacketInfo *)
						audioParser[id]->audio_stream_blocks.MakeArray();
					parentPtr->apackets[id] = audioParser[id]->audio_stream_blocks.Length();

					parentPtr->audio_sample_list[id] = (MPEGSampleInfo *)
						audioParser[id]->audio_stream_samples.MakeArray();
					parentPtr->aframes[id] = audioParser[id]->audio_stream_samples.Length();
				}
			}
		}
	}

    parentPtr->video_sample_list = (MPEGSampleInfo *)
        videoParser->video_stream_samples.MakeArray();
    parentPtr->vframes = videoParser->video_stream_samples.Length();

	DoFixUps();

	// Transfer the rest of the info
	parentPtr->width = videoParser->width;
	parentPtr->height = videoParser->height;
	parentPtr->mFrameRate = videoParser->mFrameRate;
	parentPtr->seq_ext = videoParser->seq_ext;
	parentPtr->progressive_sequence = videoParser->progressive_sequence;
	parentPtr->aspect_ratio = videoParser->aspect_ratio_info;

	parentPtr->fInterleaved = this->fInterleaved;
	parentPtr->lFirstCode = this->lFirstCode;
}

bool MPEGFileParser::guiDlgMessageLoop(HWND hDlg) {
	MSG msg;

	while(PeekMessage(&msg, NULL, 0, 0, PM_REMOVE)) {
		if (msg.message == WM_QUIT) return false;

		if (!hDlg || !IsWindow(hDlg) || !IsDialogMessage(hDlg, &msg)) {
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	}

	return true;
}

INT_PTR CALLBACK MPEGFileParser::ParseDialogProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam) {
	MPEGFileParser *thisPtr = (MPEGFileParser *)GetWindowLongPtr(hDlg, DWLP_USER);
	char buf[64];

	switch(uMsg) {

	case WM_INITDIALOG:
		{
			SetWindowLongPtr(hDlg, DWLP_USER, lParam);
			thisPtr = (MPEGFileParser *)lParam;

	// Important note:
	// The value of "fIsMPEG2" depends on "fInterleaved."
	// If fInterleaved is true, fIsMPEG2 refers to the system layer
	// If fInterleaved is false, fIsMPEG2 refers to the video stream

			if (thisPtr->fInterleaved) {
				sprintf(buf, "Parsing interleaved MPEG-%ld file",
					(thisPtr->fIsMPEG2)? 2: 1);
			} else {
				sprintf(buf, "Parsing MPEG-%ld video file",
					(thisPtr->fIsMPEG2)? 2: 1);
			}
			SetDlgItemText(hDlg, IDC_STATIC_MESSAGE, (LPSTR)buf);

			SendMessage(GetDlgItem(hDlg, IDC_PROGRESS), PBM_SETRANGE, 0, MAKELPARAM(0, 16384));
			SetTimer(hDlg, 1, 250, (TIMERPROC)NULL);
			ShowWindow(hDlg, SW_SHOW);

			return TRUE;
		}

	case WM_TIMER:
		{
			__int64 file_len = thisPtr->mFile->inSize();
			__int64 file_cpos = thisPtr->mFile->inPos();

			SendMessage(GetDlgItem(hDlg, IDC_PROGRESS), PBM_SETPOS,
				(WPARAM)((file_cpos * 16384) / file_len), 0);

			sprintf(buf, "%ldK of %ldK",
				(long)(file_cpos >> 10),
				(long)((file_len + 512) >> 10));

			SendDlgItemMessage(hDlg, IDC_CURRENT_VALUE, WM_SETTEXT, 0, (LPARAM)buf);
		}
		return TRUE;

	case WM_COMMAND:
		if (LOWORD(wParam) == IDCANCEL) thisPtr->fAbort = true;

		return TRUE;
	}

	return FALSE;
}


IMPEGFileParser *CreateMPEGFileParser(InputFileMPEG2 *const pp) {
	return new MPEGFileParser(pp);
}
